import sys
import os
import signal
import syslog
import time
import traceback
from optparse import OptionParser
from os.path import *

__all__ = ['options', 'log', 'init_options', 'init_daemon', 'cleanup']

options = OptionParser()
_got_signal = False

debug = None
info = None
err = None

def readpid(pidf, daemon):
    daemon = realpath(daemon)
    try:
        pid = int(file(pidf, 'r').readline().strip())
        #print(pid, f)
        cmdline = file('/proc/%d/cmdline' % pid, 'r').read().split('\0')
        for p in cmdline[:3]:
            if realpath(p) == daemon:
                return pid
    except Exception:
        pass
    return 0

pidf = None
    
def exitsig(sig, frame):
    global _got_signal
    if not _got_signal:
        _got_signal = True
        raise KeyboardInterrupt

def init_options():
    options.add_option("-p", "--pidfile", action="store", 
                       dest="pidfile", metavar="FILE",
                       help="write daemon process ID to a file")

    options.add_option("-l", "--logfile", action="store", 
                       dest="logfile", metavar="FILE",
                       help="redirect output to log file")

    options.add_option("-d", "--daemonize", action="store_true",
                       dest="daemon", default=False,
                       help="detach from parent")

    return options

class SyslogFile(object):
    def __init__(self, name, level=syslog.LOG_INFO, backfile=None):
        self.name = name
        self.level = level
        self.backfile = backfile
        self.lbuf = ''

    def write(self, dat):
        #ostdout.write('%r\n' % dat)
        bf = self.backfile
        ctime = time.strftime('%Y-%m-%d %H:%M:%S',
                              time.localtime(time.time()))
        prefix = '%s %s: ' % (ctime, self.name)
        bfil = self.backfile
        
        buf = self.lbuf + dat
        lines = buf.split('\n')
        for l in lines[:-1]:
            syslog.syslog(self.level, l)
            if bf:
                bf.write(prefix + l + '\n')
        self.lbuf = lines[-1]
        if bf:
            bf.flush()

    def flush(self): pass
    def close(self): pass
    def seek(self, *args): raise NotImplementedError()
    def read(self, *args): return ''

class DummyFile(object):
    def write(self, *args): pass
    def flush(self): pass
    def close(self): pass

def daemonize():
    """Makes the process into a daemon. Forks twice (to detach from
    the parent process), then calls setsid(). Closes sys.stdin,
    and redirects sys.stdout and sys.stderr to logfile, or to a dummy
    file if logfile is None.
    """

    global _logf
    
    if os.fork() != 0: os._exit(0)
    if os.fork() != 0: os._exit(0)
    os.setsid()


def write_pidfile(pidfile):
    global pidf
    try:
        pf = open(pidfile, 'w')
        pf.write('%d\n' % os.getpid())
        pf.close()
        pidf = pidfile
    except Exception:
        traceback.print_exc()
        pass
    
def init_logging(name, logfile = None, close_out = True):
    global debug, info, err
    syslog.openlog(name, syslog.LOG_NDELAY | syslog.LOG_PID,
                   syslog.LOG_DAEMON)

    if close_out:
        sys.stdout.close()
        sys.stderr.close()
        sys.stdin.close()

        logfil = None
    else:
        logfil = sys.stdout

    if logfile:
        try:
            logfil = open(logfile, "a")
        except IOError as e:
            print(('could not open logfile: %s' % e))

            
    debug = SyslogFile(name, syslog.LOG_DEBUG, logfil)
    info = SyslogFile(name, syslog.LOG_INFO, logfil)
    err = SyslogFile(name, syslog.LOG_ERR, logfil)

    sys.stdout = debug
    sys.stderr = info
    
def init_signals():
    signal.signal(signal.SIGINT, exitsig)
    signal.signal(signal.SIGTERM, exitsig)
    
def init_daemon(name, opts=None):
    if opts is None:
        opts, args = options.parse_args()

    if opts.daemon:
        daemonize()

    init_logging(name, opts.logfile, opts.daemon)
    
    if opts.pidfile:
        write_pidfile(opts.pidfile)
        
    init_signals()
    
def cleanup():
    if pidf:
        try:
            os.unlink(pidf)
        except Exception:
            pass



#!/usr/bin/python3
import sys
import re
import time
import os
import argparse
import subprocess
import traceback
import signal
import asyncio
from pathlib import Path

def run_firewall(script, scriptargs=''):
    p = argparse.ArgumentParser(description='')

    p.add_argument('-i', '--install-service',
                   nargs='?', metavar='PATH', const='/lib/systemd/system',
                   help='Install as a systemd service')

    p.add_argument('-c', '--config',
                   metavar='PATH',
                   default='/etc/firewall',
                   help='config directory')

    p.add_argument('-s', '--controlsock', dest='controlpath',
                   action='store', metavar='PATH', type=Path,
                   default=Path('/var/run/firewall-control'),
                   help='admin control socket')

    p.add_argument('-u', '--ucontrolsock', dest='ucontrolpath',
                   action='store', metavar='PATH', type=Path,
                   default=Path('/var/run/firewall-user-control'),
                   help='user control socket')

    args = p.parse_args()


    if args.install_service:
        from firewall import service
        service.install_unit(
            Path(args.install_service),
            'firewall',
            'Firewall Manager',
            '',
            Path(script).resolve(),
            scriptargs
        )
        return

    from firewall import core

    notify = None
    try:
        from systemd import daemon
        notify = lambda: daemon.notify('READY=1')
    except ImportError:
        pass

    try:
        print('firewall daemon started', file=sys.stderr)

        asyncio.run(core.run(args.config, args.controlpath, args.ucontrolpath, notify))

    except KeyboardInterrupt:
        pass
    finally:
        print('shutting down', file=sys.stderr)

if __name__ == '__main__':
    run_firewall(sys.executable, '-m firewall.main')



import sys
import re
import array
import struct
import traceback
import asyncio
import subprocess

import os
import socket
import fcntl

from io import StringIO

from firewall import util, command
from firewall.core import *

public = util.public(globals())

networks = {}
_iptables_running = False

addrstruct = struct.Struct(">I")

class NetworkUnavail(Exception):
    pass
public(NetworkUnavail)


@public
def pack_addr(str):
    if str is None:
        return None
    try:
        return addrstruct.unpack(socket.inet_aton(str))[0] 
    except socket.error:
        return 0

@public
def unpack_addr(addr):
    return '.'.join(str((addr >> i) & 255) for i in range(24, -8, -8))

def _get_iface_info(iface, ctl):
    buf = bytearray(40)
    iface = bytes(iface, 'utf-8')
    buf[:len(iface)] = iface
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        saddr = None
        try:
            fd = s.fileno()
            res = fcntl.ioctl(fd, ctl, buf)
            return struct.unpack_from(">I", buf, 20)[0]
        except OSError:
            raise NetworkUnavail(iface) from None

class Network(object):
    public = True
    def __init__(self, iface=None):
        self.iface = iface

    def _getaddr(self):
        pass

    def _getmask(self):
        pass

    def getaddr(self):
        return unpack_addr(self._getaddr())

    def getmask(self):
        return unpack_addr(self._getmask())

    def getnetaddr(self):
        addr = self._getaddr()
        mask = self._getmask()
        return unpack_addr(addr & mask)

    def getbcastaddr(self):
        addr = self._getaddr()
        mask = self._getmask()
        return unpack_addr(addr | ~mask)

    def getnet(self):
        addr = self._getaddr()
        mask = self._getmask()
        return '%s/%s' % (unpack_addr(addr & mask), unpack_addr(mask))

    def getiface(self):
        return (self.iface or '<none>')
    
class StaticNetwork(Network):
    public = True
    def __init__(self, addr, mask, iface=None):
        Network.__init__(self, iface)
        self.addr = addr
        self.mask = mask
        
    def _getaddr(self):
        return self.addr
    
    def _getmask(self):
        return self.mask

SIOCGIFADDR = 0x8915
SIOCGIFNETMASK = 0x891b

class NetInterface(Network):
    public = True
    def __init__(self, iface):
        Network.__init__(self, iface)

    def _getaddr(self):
        return _get_iface_info(self.iface, SIOCGIFADDR)
        
    def _getmask(self):
        return _get_iface_info(self.iface, SIOCGIFNETMASK)

@public
def parse_mask(mstr):
    try:
        nbits = int(mstr)
        if nbits <= 32:
            return 0xffffffff << (32 - nbits)
    except ValueError:
        pass

    return pack_addr(str)

@public
def parse_network(str):
    if str.startswith('@'):
        return NetInterface(str[1:])
    
    lst = str.split('/', 2)
    addr = pack_addr(lst[0])
    if len(lst) == 1:
        mask = 0xffffffff
    else:
        mask = parse_mask(lst[1])
        
    if len(lst) == 3:
        iface = lst[2]
    else:
        iface = None
    return StaticNetwork(addr, mask, iface)

@public
def get_network(nm):
    try:
        return networks[nm]
    except KeyError:
        return parse_network(nm)
    
@public
def get_network_list(str):
    return [get_network(nm) for nm in str.split(' ') if nm]

@public
def parsetime(t):
    lst = t.split(':',2)
    secs = 0.0
    for s in lst:
        secs *= 60
        if s:
            secs += float(s)
    return secs
    
class Rule(object):
    public = True
    def __init__(self, match='', action=''):
        self.match = match
        self.action = action

    def format(self):
        return '-j %s %s'%(self.action, self.match)

class Chain(object):
    public = True
    def __init__(self, name, policy='-'):
        self.name = name
        self.policy = policy
        self.rules = []

    def add(self, rule):
        self.rules.append(rule)

    def clear(self):
        self.rules = []

    def format(self):
        return ''.join('-A %s %s\n'%(self.name, r.format()) for r in self.rules)

class Table(object):
    public = True
    def __init__(self, name):
        self.name = name
        self.chains = {}

    def __getitem__(self, item):
        try:
            return self.chains[item]
        except KeyError:
            ret = Chain(item)
            self.chains[item] = ret
            return ret
        
    def __setitem__(self, item, value):
        self.chains[item] = value

    def format(self):
        lins = ['*%s\n' % self.name]
        
        for k, v in self.chains.items():
            lins.append(':%s %s [0:0]\n'%(k, v.policy))
        
        for k, v in self.chains.items():
            lins.append(v.format())
        lins.append('COMMIT\n')
        return ''.join(lins)

_reload_task = None

@public
async def reload_tables():
    global _reload_task
    initial_task = _reload_task

    if initial_task is not None and not initial_task.done():
        try:
            await initial_task
        except Exception:
            pass

    if initial_task is _reload_task:
        _reload_task = asyncio.create_task(_reload_tables())

    return await _reload_task

@public
async def _reload_tables():
    raw = Table('raw')
    raw['PREROUTING'] = Chain('PREROUTING', 'ACCEPT')
    raw['OUTPUT'] = Chain('OUTPUT', 'ACCEPT')

    filter = Table('filter')
    filterin = filter['INPUT'] = Chain('INPUT', 'ACCEPT')
    filterout = filter['OUTPUT'] = Chain('OUTPUT', 'ACCEPT')
    filterfwd = filter['FORWARD'] = Chain('FORWARD', 'ACCEPT')

    mangle = Table('mangle')
    mangle['INPUT'] = Chain('INPUT', 'ACCEPT')
    mangle['OUTPUT'] = Chain('OUTPUT', 'ACCEPT')
    mangle['FORWARD'] = Chain('FORWARD', 'ACCEPT')
    mangle['POSTROUTING'] = Chain('POSTROUTING', 'ACCEPT')
    mangle['PREROUTING'] = Chain('PREROUTING', 'ACCEPT')

    nat = Table('nat')
    natout = nat['OUTPUT'] = Chain('OUTPUT', 'ACCEPT')
    natpost = nat['POSTROUTING'] = Chain('POSTROUTING', 'ACCEPT')
    natpre = nat['PREROUTING'] = Chain('PREROUTING', 'ACCEPT')

    await callhook('fill_tables', filter, nat, mangle, raw)

    txt = raw.format() + mangle.format() + filter.format() + nat.format()

    try:
        proc = await asyncio.create_subprocess_exec(iptables_bin, stdin=subprocess.PIPE)
        await proc.communicate(txt.encode('utf8'))
        ret = await proc.wait()
    except (IOError, OSError):
        traceback.print_exc()
        ret = 1

    if ret:
        print("iptables-restore exited with code %d" % ret, file=sys.stderr)
    else:
        print("iptables loaded successfully", file=sys.stderr)

    return ret

@public
async def reload_tables_for_cmd(source, stxt):
    if source != SOURCE_CONFIG:
        stat = await reload_tables()
        if not stat:
            return (command.RESP_OK, stxt)
        else:
            return (stat, 'iptables error')

    return (command.RESP_OK, stxt)

def clear_config():
    global iptables_bin
    iptables_bin = '/sbin/iptables-restore'
    networks.clear()


@admin_command
async def command_reload(source):
    return await reload_tables_for_cmd(source, 'iptables loaded successfully')

@admin_command
async def command_network(source, name, val):
    networks[name] = get_network(val)
    return (command.RESP_OK, 'network defined')

@admin_command
async def command_iptables_bin(source, bin):
    global iptables_bin
    iptables_bin = bin
    
    return await reload_tables_for_cmd(source, 'ok')

@user_command
def listnets(source):
    ret = StringIO()
    for k, v in networks.items():
        print('%s = %s %s %s' % (k.ljust(15), v.getiface().ljust(8), v.getaddr().ljust(20), v.getnet()), file=ret)
    return (0, ret.getvalue())

async def config_loaded():
    return await reload_tables()

def init():
    clear_config()

def cleanup():
    pass

util.update(globals())



import sys
import re
import asyncio

from io import StringIO

from firewall.core import *
from firewall.modules.iptables import *

pinholes = {}
protect = {}

class Pinhole(object):
    def __init__(self, rulesrc, proto='', portrange='', source='', interface=''):
        self.rulesrc = rulesrc

        if (portrange and not proto) or proto == 't':
            proto = 'tcp'
        elif proto == 'u':
            proto = 'udp'

        self.proto = proto
        self.portrange = portrange
        self.source = source
        self.interface = interface
        self.active = True

    def ident(self):
        return (self.proto, self.portrange, self.source, self.interface)

    def __str__(self):
        return ':'.join(self.ident())

    def genrule(self):
        if self.proto:
            ruletxt = '-m %s -p %s' % (self.proto, self.proto)
        else:
            ruletxt = ''

        if self.portrange:
            lst = self.portrange.split('-')
            if len(lst) == 1:
                ruletxt += ' --dport %s' % lst[0]
            else:
                ruletxt += ' --dport %s:%s' % (lst[0], lst[1])

        if self.source:
            ruletxt += ' --source %s' % get_network(self.source).getnet()

        if self.interface:
            ruletxt += ' --in-interface %s' % get_network(self.interface).getiface()

        return Rule(ruletxt, 'RETURN')

class Protect(object):
    def __init__(self, rulesrc, ifaces, nets):
        self.rulesrc = rulesrc
        self.ifaces = tuple(sorted(ifaces))
        self.nets = nets

    def ident(self):
        return self.ifaces

    def gen_table_name(self):
        return 'from_' + '_'.join(self.ifaces)

def fill_tables(filter, nat, mangle, raw):
    filterin = filter['INPUT']
    filterout = filter['OUTPUT']
    filterfwd = filter['FORWARD']
    servers = filter['SERVERS']

    rawpre = raw['PREROUTING']

    filterin.add(Rule('-p tcp -m tcp --tcp-flags SYN,RST,ACK SYN', 'SERVERS'))
    filterin.add(Rule('-p udp -m udp -m conntrack --ctstate NEW', 'SERVERS'))

    servers.add(Rule('-i lo', "RETURN"))

    for p in list(pinholes.values()):
        if p.active:
            try:
                servers.add(p.genrule())
            except NetworkUnavail as e:
                pass
                #print >> sys.stderr, "interface not found: %s" % e

    for p in list(protect.values()):
        tblname = p.gen_table_name()
        tbl = raw[tblname]
        for iface in p.ifaces:
            try:
                rawpre.add(Rule('--in-interface %s' % get_network(iface).getiface(), tblname))
            except NetworkUnavail as e:
                pass
                #print >> sys.stderr, "interface not found: %s" % e

        tbl.add(Rule('-m conntrack --ctstate ESTABLISHED', 'RETURN'))
        for net in p.nets:
            try:
                tbl.add(Rule('--destination %s' % get_network(net).getnet(), 'DROP'))
            except NetworkUnavail as e:
                pass
                #print >> sys.stderr, "interface not found: %s" % e
    servers.add(Rule('', 'DROP'))

@admin_command
async def pin(source, *args):
    added = []
    for v in args:
        lst = v.split(':')
        ph = Pinhole(source, *lst)
        ph = pinholes.setdefault(ph.ident(), ph)
        added.append(str(ph))

    return await reload_tables_for_cmd(source, 'added pinholes: %s' % ','.join(added))

@admin_command
async def rmpin(source, *args):
    deleted = []
    for v in args:
        v = str.split(':')
        ph = Pinhole(source, *args)
        try:
            del pinholes[ph.ident()]
            deleted.append(str(ph))
        except KeyError:
            pass


    return await reload_tables_for_cmd(source, 'deleted pinholes: %s' % ','.join(deleted))

@admin_command
async def flushtemp(source):
    deleted = []
    for k, ph in list(pinholes.items()):
        if ph.source == SOURCE_ADMIN:
            del pinholes[k]
            deleted.append(str(ph))
    return await reload_tables_for_cmd(source, 'deleted pinholes: %s' % ','.join(deleted))


@admin_command
async def lspin(source):
    txt = StringIO()
    for k, ph in list(pinholes.items()):

        txt.write('%s/%s' % (ph.portrange or '<any>', ph.proto or '<any>'))
        if ph.interface:
            txt.write(', from iface %s' % ph.interface)

        if ph.source:
            txt.write(', from source %s' % ph.source)
        txt.write('\n')
    return (0, txt.getvalue())

@admin_command
async def command_protect(source, *args):
    ifaces = []
    nets = []
    i = iter(args)
    for v in i:
        if v == 'from':
            break
        nets.append(v)
    for v in i:
        ifaces.append(v)

    prot = Protect(source, ifaces, [])
    prot = protect.setdefault(prot.ident, prot)
    prot.nets.extend(nets)

    return await reload_tables_for_cmd(source, 'added protect')

def clear_config():
    protect.clear()
    for k, ph in list(pinholes.items()):
        if ph.source == SOURCE_CONFIG:
            del pinholes[k]

def cleanup():
    pass

def init():
    pass



import sys
import re
import asyncio

from io import StringIO

from firewall.core import *
from firewall.modules.iptables import *

class ManualRule(object):
    def __init__(self, table, chain, ruleargs, target):
        self.table = table
        self.chain = chain
        self.ruleargs = ruleargs
        self.target = target

rules = []


def filter_args(seq):
    for i in seq:
        if i.startswith('i:'):
            yield get_network(i[2:]).getiface()
        elif i.startswith('a:'):
            yield get_network(i[2:]).getaddr()
        elif i.startswith('m:'):
            yield get_network(i[2:]).getmask()
        elif i.startswith('b:'):
            yield get_network(i[2:]).getbcast()
        elif i.startswith('n:'):
            yield get_network(i[2:]).getnet()
        else:
            yield i

def fill_tables(filter, nat, mangle, raw):
    tbls = dict(filter = filter,
                nat = nat,
                mangle = mangle,
                raw = raw)

    for r in rules:
        try:
            ruletxt = ' '.join(filter_args(r.ruleargs))
            tbls[r.table][r.chain].add(Rule(ruletxt, r.target))
        except NetworkUnavail:
            pass
        except Exception as exc:
            print(str(exc), file=sys.stderr)

@admin_command
async def rule(source, table, chain, target, *args):
    #ruletxt = ' '.join((get_network(arg[1:]) if arg.startswith('$') else arg) for arg in args)
    rules.append(ManualRule(table, chain, args, target))
    return await reload_tables_for_cmd(source, 'ok')

def clear_config():
    del rules[:]

def init():
    pass

def cleanup():
    pass




import sys
import traceback
import time
import socket
import asyncio

from io import BytesIO, StringIO


def httptime(tm = None):
    if tm is None:
        tm = time.time()
    return time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime(tm))

def msgpage(txt, extra=''):
    return '<html><head><title>%s</title></head><body>\n<h1>%s</h1>\n%s\n</body></html>'%(txt, txt, extra)

# Bare minimal http server.
# I use this instead of BaseHTTPServer because BaseHTTPServer does not
# support async I/O.

class HTTPRequest(object):
    def __init__(self, reader, writer):
        self.uri = ''
        self.headers_in = {}
        self.headers_out = {}
        self.body_buf = BytesIO()
        self.status = 200
        self.msg = 'OK'
        self.headers_out['content-type'] = 'text/html'
        self.reader = reader
        self.writer = writer
        
    def write(self, data):
        if isinstance(data, str):
            data = data.encode('utf8')
        self.body_buf.write(data)
        
    async def simpleerror(self, status, msg):
        self.write(msgpage(msg))
        return await self.send_response(status, msg)
        
    async def send_response(self, status=None, msg=None):
        response = StringIO()
        if status is None:
            status = self.status
        if msg is None:
            msg = self.msg

        body = self.body_buf.getvalue()
        self.putheader('date', httptime())
        self.putheader('content-length', str(len(body)))

        self.writer.write(('HTTP/1.1 %03d %s\r\n' % (status, msg)).encode('ascii'))
        
        for k, v in self.headers_out.items():
            k = k.title()
            lst = v.split('\n')
            for i in lst:
                self.writer.write(('%s: %s\r\n' % (k, i)).encode('utf8'))
        
        self.writer.write(b'\r\n')
        self.writer.write(body)
        await self.writer.drain()
        self.writer.close()
        await self.writer.wait_closed()

    def nocache(self):
        self.headers_out['cache-control'] = 'no-cache'
        self.headers_out['pragma'] = 'no-cache'
        self.headers_out['expires'] = httptime()
        
    def getheader(self, hdr, default=''):
        return self.headers_in.get(hdr, default)

    def putheader(self, hdr, value):
        self.headers_out[hdr] = value

    def getpeername(self):
        return self.writer.get_extra_info('peername')
    
    def redirect(self, loc, perm = False):
        self.putheader('location', loc)
        self.write(msgpage('Detour', '<a href="%s">linky</a>' % loc))
        return self.send_response((301 if perm else 302), 'Taking a detour')

async def handle_request(reader, writer, processor):
    req = HTTPRequest(reader, writer)

    reqline = (await reader.readline()).decode('utf8', 'ignore').rstrip('\r\n')
    reqd = reqline.split(' ')
    if len(reqd) != 3:
        return await req.simpleerror(400, 'Bad request')
    
    method = reqd[0]
    if method != 'GET':
        return await req.simpleerror(400, 'Unsupported method')
    
    req.uri = reqd[1]

    while True:
        reqline = (await reader.readline()).decode('utf8', 'ignore').rstrip('\r\n')
        if not reqline:
            break
        lst = reqline.split(':', 1)
        if len(lst) != 2:
            yield req.simpleerror(400, 'Malformed header')
            return
        req.headers_in[lst[0].lower()] = lst[1].strip()
        
    try:    
        await processor(req)
    except Exception as exc:
        req.write(msgpage('Sorry'))
        # Run it asynchronously so we can propagate the error
        asyncio.create_task(req.send_response(500, 'error'))
        raise
    
async def http_server(addr, taskfunc):
    return await asyncio.start_server(lambda reader, writer: handle_request(reader, writer, taskfunc), host=addr[0], port=addr[1])
    



import os
from pathlib import Path


SERVICE_TEMPLATE='''\
[Unit]
Description={name}
Before=network.target
{after}

[Service]
Type=notify
WorkingDirectory=/
ExecStart={cmd}
Restart=always

[Install]
WantedBy=network.target
'''

def install_unit(basepath, unit, name, after, cmd, args):
    cmd = str(cmd)
    if ' ' in cmd:
        cmd = '"' + cmd + '"'

    if args:
        cmd += ' ' + args

    text = SERVICE_TEMPLATE.format(
        name=name,
        after=after,
        cmd=cmd,
        args=args)
    svcfile = basepath / f'{unit}.service'
    svcfile.write_text(text, encoding='utf8')
    try:
        os.chmod(svcfile, 0o644)
        os.chown(svcfile, 0, 0)
    except OSError:
        pass



from ctypes import *
import socket
from sys import platform

__all__ = ['NetAddr',
           'getifaddrs', 'addr_to_long', 'long_to_addr']
 
class ifaddrs_ifa_ifu(Union):
    _fields_ = [ 
        ("ifu_broadaddr", c_void_p),
        ("ifu_dstaddr",   c_void_p)  
        ]
    
class ifaddrs(Structure):
    _fields_ = [
        ("ifa_next",    c_void_p),
        ("ifa_name",    c_char_p),
        ("ifa_flags",   c_uint),
        ("ifa_addr",    c_void_p),
        ("ifa_netmask", c_void_p),
        ("ifa_ifu",     ifaddrs_ifa_ifu),
        ("ifa_data",    c_void_p)
        ]
class sockaddr(Structure):
    _fields_ = [
        ("sa_family", c_uint16),
        ("sa_data",   (c_uint8 * 14) ) 
        ]

class in_addr(Union):
    _fields_ = [
        ("s_addrstr",  (c_char * 4) ),
        ("s_addr", c_uint32),
    ]
    
class sockaddr_in(Structure):
    _fields_ = [
        ("sin_family", c_short),
        ("sin_port",   c_ushort),
        ("sin_addr",   in_addr),
        ("sin_zero",   (c_char * 8) ), # padding
        ]
    
class in6_addr(Union):
    _fields_ = [
        ("s6_addr",  (c_uint8 * 16) ),
        ("s6_addrstr", (c_char * 16) ),
        ]
    
class sockaddr_in6(Structure):
        _fields_ = [
            ("sin6_family",   c_short),
            ("sin6_port",     c_ushort),
            ("sin6_flowinfo", c_uint32),
            ("sin6_addr",     in6_addr),
            ("sin6_scope_id", c_uint32),
        ]

class NetAddr(object):
    def __init__(self, iface=None, family=0, addr=0, mask=0, scope=0):
        self.family = family
        
        self.iface = iface
        
        if isinstance(addr, str):
            addr = addr_to_long(self.family, addr)
            
        if isinstance(mask, str):
            mask = addr_to_long(self.family, mask)
        
        self.addr = addr
        self.mask = mask
        bits = 128
        while bits and not (mask & 1):
            bits -= 1
            mask >>= 1
        self.bits = bits
        self.scope = scope

    def getaddr(self):
        return long_to_addr(self.family, self.addr)

    def getmask(self):
        return long_to_addr(self.family, self.mask)

    def getnetaddr(self):
        return long_to_addr(self.family, self.addr & self.mask)

    def getbcastaddr(self):
        return long_to_addr(self.family, self.addr | ~self.mask)

    def getnet(self):
        return '%s/%d' % (long_to_addr(self.family, self.addr & self.mask), self.bits)

    def getiface(self):
        return (self.iface or '<none>')

    def copy(self, iface=None, addr=None, mask=None, scope=None):
        if iface is None:
            iface = self.iface
        if addr is None:
            addr = self.addr
        if mask is None:
            mask = self.mask
        if scope is None:
            scope = self.scope
        return NetAddr(iface, self.family, addr, mask, scope)

    def __repr__(self):
        return 'NetAddr(%r, %d, %r, %r, %r)' % (self.iface, self.family, self.getaddr(), self.getmask(), self.scope)

def pstr_to_long(s):
    addr = 0
    for c in str(s):
        addr <<= 8
        addr += ord(c)
    return addr

def addr_to_long(family, str):
    try:
        paddr = socket.inet_pton(family, str)
    except socket.error:
        raise ValueError('invalid IP address: %r' % str)
    return pstr_to_long(paddr)

def long_to_addr(family, addr):
    bits = 32
    if family == socket.AF_INET6:
        bits = 128
    paddr = ''.join(chr((addr >> i) & 255) for i in range(bits - 8, -8, -8))
    #print family, addr, repr(paddr)
    return socket.inet_ntop(family, paddr)


def get_raw_data(cto):
    return string_at(addressof(cto), sizeof(cto))
    

libc = CDLL("libc.so.6")

def getifaddrs():
    ptr = c_void_p(None)
    if libc.getifaddrs(pointer(ptr)):
        return
    
    ifa = ifaddrs.from_address(ptr.value)
    result = {}
 
    while True:
        name = ifa.ifa_name

        iface = result.get(name)
        if iface is None:
            iface = result[name] = []
        if ifa.ifa_addr:
            sa = sockaddr.from_address(ifa.ifa_addr)
            family = sa.sa_family
             
        addrobj = None
 
        if family == socket.AF_INET:
            addr = 0
            netmask = 0
            
            if ifa.ifa_addr is not None:
                si = sockaddr_in.from_address(ifa.ifa_addr)
                addr = pstr_to_long(get_raw_data(si.sin_addr))
                
            if ifa.ifa_netmask is not None:
                si = sockaddr_in.from_address(ifa.ifa_netmask)
                netmask = pstr_to_long(get_raw_data(si.sin_addr))
            addrobj = NetAddr(name, socket.AF_INET, addr, netmask)
                
        elif family == socket.AF_INET6:
            addr = 0
            netmask = 0
            scope = 0
            if ifa.ifa_addr is not None:
                si = sockaddr_in6.from_address(ifa.ifa_addr)
                addr = pstr_to_long(get_raw_data(si.sin6_addr.s6_addr))
                
                if (addr >> 112) & 0xff80 == 0xfe80:
                    scope = si.sin6_scope_id
                    
            if ifa.ifa_netmask is not None:
                si = sockaddr_in6.from_address(ifa.ifa_netmask)
                netmask = pstr_to_long(get_raw_data(si.sin6_addr.s6_addr))
 
            addrobj = NetAddr(name, socket.AF_INET6, addr, netmask, scope)
 
        if addrobj:
            iface.append(addrobj)
 
        if ifa.ifa_next:
            ifa = ifaddrs.from_address(ifa.ifa_next)
        else:
            break
 
    libc.freeifaddrs(ptr)
    return result

if __name__ == '__main__':
    print(getifaddrs())




import sys
import re
import os
import time
import struct
from os.path import *

BLOCKSHIFT = 12
BLOCKSIZE = 1 << BLOCKSHIFT

def gettime():
    return int(time.time() * 1000)

def readvar(str, pos):
    ret = 0
    while True:
        rd = ord(str[pos])
        pos += 1
        ret <<= 7
        ret |= rd & 127
        if not rd & 128:
            return ret, pos

def encodev(v):
    if v < 128:
        return chr(v)
    if v < 16384:
        return chr(0x80 | ((v >> 7) & 0x7F)) + chr(v & 0x7F)
    if v < 2097152:
        return chr(0x80 | ((v >> 14) & 0x7F)) + chr(0x80 | ((v >> 7) & 0x7F)) + chr(v & 0x7F)

    r = ''
    while v:
        tv = v & 0x7F
        v >>= 7
        if r:
            tv |= 0x80
        r = chr(tv) + r
        
    return r


class Counter(object):
    def __init__(self, v=0, name=None):
        self.v = v
        self.delta = 0
        self.name = name

    def getdelta(self):
        delt = self.delta
        self.v += delt
        self.delta = 0
        return delt
        
DBG = True
MAX_TIME = 1 << 64

iotime = 0
strll = struct.Struct('>Q')
class TrafLogReader(object):
    def __init__(self, fn, ctrs=None):
        self.fn = fn
        self.f = open(fn, 'rb')
        self.lastlen = None
        self.check_update()

        self.cblock = self.nblocks
        self.blockdat = None
        self.bstime = MAX_TIME

        self.bpos = 0
        self.ltime = 0

        self.blocktimecache = {}
        #self.blocktimecache = [None] * self.nblocks
        if ctrs is None:
            ctrs = []
        self.ctrs = ctrs

    def close(self):
        if self.f:
            self.f.close()
            self.f = None

    def blocktime(self, n):
        bt = self.blocktimecache.get(n)
        if bt is not None:
            return bt

        bt = self.blocktimecache[n] = self.seekblock(n)
        return bt

    def seekblock(self, n):
        if n < 0:
            n += self.nblocks

        self.f.seek(n << BLOCKSHIFT)
        self.blockdat = b = self.f.read(BLOCKSIZE)

        self.cblock = n
        self.bstime = sttime = strll.unpack(b[:8])[0]
        self.bpos = 0

        return sttime

    def _seek_block_begin(self):
        b = self.blockdat
        ctrs = self.ctrs
        nctrs = ord(b[8])
        while len(ctrs) < nctrs:
            ctrs.append(Counter())

        pos = 9
        for i in range(nctrs):
            v = strll.unpack(b[pos : pos + 8])[0]
            ctrs[i].v = v
            pos += 8

        self.ltime = self.bstime
        self.bpos = pos

    def check_update(self):
        self.f.seek(0, 2)
        flen = self.f.tell()
        nblocks = (flen + BLOCKSIZE - 1) >> BLOCKSHIFT
        if self.lastlen is not None:
            if flen > self.lastlen:
                if self.cblock == self.nblocks - 1:
                    # if current block is last block, re-read it
                    self.seekblock(self.cblock)
        self.nblocks = nblocks
        self.lastlen = flen

    def seektime(self, t):
        #if t == self.ltime:
        #    return True

        #if DBG: print 'seektime %x' % t
        if not self.blockdat:
            self.binseek(t, 0, self.nblocks)

        elif t >= self.bstime:
            #if DBG: print '  > bstime %x' % self.bstime
            if t < self.ltime:
                #if DBG: print '  < ltime %x' % self.ltime
                self._seek_block_begin()

            if self._parse_block_fwd(t):
                return True
            #if DBG: print '  > last time in block'
            self.binseek(t, self.cblock, self.nblocks)
            
        else:
            #if DBG: print '  < bstime %x' % self.bstime
            self.binseek(t, 0, self.cblock)
            if t < self.bstime:
                #if DBG: print 'out of range 1'
                return False

        self._seek_block_begin()
        return self._parse_block_fwd(t) or self.cblock != self.nblocks - 1

    def seeknext(self):
        if not self._parse_block_fwd(None, True):
            if self.cblock == self.nblocks - 1:
                return False
            self.seekblock(self.cblock + 1)
            self._seek_block_begin()
        return True

    def seek_begin(self):
        self.seekblock(0)
        self._seek_block_begin()

    def seek_end(self):
        self.seekblock(-1)
        self._seek_block_begin()
        self._parse_block_fwd(None)
        
    def binseek(self, seektm, bsb, bse):
        #if DBG: print '    binseek seektm=%x bsb=%d bse=%d' % (seektm, bsb, bse)
        safety = 10000
        while safety:
            med = (bsb + bse) // 2
            #if DBG: print '    binseek bsb=%d bse=%d med=%d' % (bsb, bse, med)
            if bsb == bse:
                break
            
            cbt = self.blocktime(med)
            #if DBG: print '    binseek cbt=%x' % (cbt)

            if cbt <= seektm:
                if bsb == med:
                    break
                bsb = med

            else:
                if bse == med:
                    med -= 1
                    break
                bse = med

            safety -= 1

        #if DBG: print '    binseek med=%d' % (med)
        if med != self.cblock or med == self.nblocks - 1:
            self.seekblock(med)
        

    def _parse_block_fwd(self, t, next=False):
        b = self.blockdat
        nctrs = ord(b[8])
        pos = self.bpos
        ltime = ctime = self.ltime
        ln = len(b)
        ctrs = self.ctrs
        
        if len(ctrs) > nctrs:
            ctrs = ctrs[:nctrs]
        #if DBG: print '    _pbf ctime=%x pos=%d ln=%d' % (ctime, pos, ln)
        try:
            while pos < ln:
                #if DBG: print '      pos=%d' % (pos)
                td, npos = readvar(b, pos)
                #if DBG: print '      td=%d' % (td)
                if not td:
                    break
                ctime += td
                #if DBG: print '      ctime=%x' % (ctime)
                
                if t is not None and ctime > t:
                    self.bpos = pos
                    self.ltime = ltime
                    return True

                for c in ctrs:
                    cd, npos = readvar(b, npos)
                    c.v += cd
                    
                if next:
                    self.bpos = npos
                    self.ltime = ctime
                    return True

                ltime = ctime
                pos = npos
        except IndexError:
            pass

        self.bpos = pos
        self.ltime = ltime
        return False


class TrafLogWriter(object):
    def __init__(self, fn, ctrs):
        self.fn = fn
        self.f = open(fn, 'a+b', buffering=0)
        flen = self.f.tell()
        pad = BLOCKSIZE - (flen & (BLOCKSIZE - 1))
        if pad < BLOCKSIZE:
            self.f.write('\0' * pad)

        self.block_pos = BLOCKSIZE
        self.ctrs = ctrs
        self.ctime = 0

    def close(self):
        if self.f:
            self.f.close()
            self.f = None

    def write_counters(self, time=None):
        if time is None:
            time = gettime()
        timedelt = time - self.ctime
        self.ctime = time
        wdata = encodev(timedelt) + ''.join(encodev(c.getdelta()) for c in self.ctrs)
        nblockpos = self.block_pos + len(wdata)

        if nblockpos > BLOCKSIZE:
            pad = BLOCKSIZE - self.block_pos
            hdr = strll.pack(time) + chr(len(self.ctrs)) + \
                ''.join(strll.pack(c.v) for c in self.ctrs)
            wdata = ('\0' * pad) + hdr
            nblockpos = len(hdr)

        self.f.write(wdata)
        self.block_pos = nblockpos

rxtimef = re.compile(r'([0-9a-f]{16})_')
class TrafLog(object):
    def __init__(self, dir, ctrs):
        self.dir = dir
        try:
            os.makedirs(dir)
        except OSError:
            pass
        self.lastmod = None
        self.cday = None
        self.checkfiles()

        self.curidx = None
        self.curtf = None
        try:
            self.ctrs = list(ctrs)
        except TypeError:
            self.ctrs = [Counter(0) for x in range(ctrs)]

    def close(self):
        if self.curtf:
            self.curtf.close()
            self.curtf = None

    def checkfiles(self):
        lm = getmtime(self.dir)
        if lm == self.lastmod:
            return

        files = self.files = []
        for f in os.listdir(self.dir):
            m = rxtimef.match(f)
            if m:
                fn = join(self.dir, f)
                if getsize(fn) >= 16:
                    ct = int(m.group(1), 16)
                    files.append((ct, fn))
        files.sort()

    def seeknext(self):
        if self.curtf:
            if self.curtf.seeknext():
                self.ltime = self.curtf.ltime
                return True
            if self.curidx == len(self.files) - 1:
                return False
            
            self.curtf.close()
            self.curidx += 1
        else:
            self.curidx = 0
        self.curtf = TrafLogReader(self.files[self.curidx][1], self.ctrs)
        self.curtf.seek_begin()
        self.ltime = self.curtf.ltime
        return True
    
    def seektime(self, t):
        if self.curtf:
            if self.curtf.seektime(t):
                return True
            self.curtf.close()
            self.curtf = None
            self.curidx = None
        lf = None
        lidx = None
        for i, (ct, f) in enumerate(self.files):
            if ct > t:
                break
            lf = f
            lidx = i
            
        if not lf:
            return False
        
        self.curtf = TrafLogReader(lf, self.ctrs)
        self.curidx = lidx
        return self.curtf.seektime(t)

    def seek_end(self):
        self.seektime(MAX_TIME)

    def genfn(self, tm):
        tstr = time.strftime('%Y_%m_%d_%H_%M_%S', time.localtime(tm / 1000))
        return '%016x_%s' % (tm, tstr)

    def write_counters(self, time=None):
        if time is None:
            time = gettime()
        day = time / 86400000
        if self.curtf is None or day != self.cday:
            self.cday = day
            if self.curtf:
                self.curtf.close()
            newfn = self.genfn(time)
            self.curtf = TrafLogWriter(join(self.dir, newfn), self.ctrs)
        self.curtf.write_counters(time)
        

def test1():
    tstlogf = 'traflog'
    
    TMUL = 5000
    c1 = Counter(15)
    c2 = Counter(63)
    print('writing...')

    tw = TrafLog(tstlogf, [c1, c2])
    for j in range(100000):
        tw.write_counters(j * TMUL)
        c1.delta = 0x100
        c2.delta = 0x1000
    tw.close()

    print('reading...')
    maxtime = TMUL * 100000
    tr = TrafLog(tstlogf, 2)

    st = time.time()
    ct = 0
    while True:
        if not tr.seektime(ct):
            break
        rtime = ct & ~0xFFFFFF
        crc = ct / TMUL
        c1b = 15 + crc * 0x100
        c2b = 63 + crc * 0x1000
        c1q = tr.ctrs[0].v
        c2q = tr.ctrs[1].v
        if c1b != c1q or c2b != c2q:
            print('xxx %x' % ct)
            print('%016x %016x' % (c1b, c1q))
            print('%016x %016x' % (c2b, c2q))
        ct += TMUL
    et = time.time()
    print('%x %.15f %.15f' % (ct, et - st, iotime))
    return 
    while True:
        st = time.time()
        trytime = random.randrange(0, maxtime)
        #trytime = 0
        tr.seektime(trytime)
        rtime = trytime & ~0xFFFFFF
        crc = trytime / TMUL
        c1b = 15 + crc * 0x100
        c2b = 63 + crc * 0x1000
        c1q = tr.ctrs[0].v
        c2q = tr.ctrs[1].v
        et = time.time()
        print('%x %.15f %.15f' % (ct, et - st, iotime))
        print('%016x %016x' % (c1b, c1q))
        print('%016x %016x' % (c2b, c2q))
        if c1b != c1q or c2b != c2q:
            print(trytime)
            break
        print()

def test2():
    tstlogf = 'traflog'

    tw = TrafLog(tstlogf, 5)
    ctrs = tw.ctrs

    tcnt = 0

    samples = []
    print('writing ...')
    ct = gettime()
    nextsample = random.randrange(20, 200)
    ctdelta = 0
    while len(samples) < 200:
        tcnt += 1
        ct += ctdelta
        ctdelta = random.randrange(300, 10000)
        for c in ctrs:
            c.delta = random.randrange(0, 10000000)
        tw.write_counters(ct)
        nextsample -= 1
        if nextsample == 0:
            samples.append((ct, ct + ctdelta, [c.v for c in ctrs]))
            nextsample = random.randrange(20, 100)
    tw.close()
    print('wrote %d entries' % tcnt)
    
    tr = TrafLog(tstlogf, 5)
    bt = time.time()
    print('reading sequential ...')
    i = 0
    j = 0
    tstnext = None
    while tr.seeknext():
        j += 1
        #if j < 10 or j % 10000 == 0:
        #    print j, tr.ltime
        if i < len(samples) and tr.ltime == samples[i][0]:
            ct, nct, ctrs = samples[i]
            i += 1
            tstnext = nct
            tcv = [c.v for c in tr.ctrs]
            if tcv != ctrs:
                print('!!!!! %d %r %r' % (ct, tcv, ctrs))
        elif tstnext is not None:
            if tr.ltime != tstnext:
                print('!!!!! %d %d %d %r %r' % (tr.ltime, tstnext, ct, tcv, ctrs))
            tstnext = None
    
    et = time.time()
    print('read %d sequential log entries in %.5fs' % (j, et - bt))
    
    print('reading ...')
    random.shuffle(samples)
    #samples = [random.choice(samples)]
    tr.seektime(0)


    bt = time.time()
    for ct, nct, ctrs in samples:
        tsttime = random.randrange(ct, nct)
        tr.seektime(tsttime)
        tcv = [c.v for c in tr.ctrs]
        if tcv != ctrs:
            print('!!!!! %d %r %r' % (tsttime, tcv, ctrs))
        #print 'ok %d' % tsttime
        #else:
        #    print '!!!!! %d %r %r' % (tsttime, tcv, ctrs)
    et = time.time()
    print('read %d random samples in %.5fs' % (len(samples), et - bt))

            
def main():
    if len(sys.argv) > 1:
        dir = sys.argv[1]
        tr = TrafLog(dir, 0)
        while tr.seeknext():
            tcv = [c.v for c in tr.ctrs]
            print('%d %r' % (tr.ltime, tcv))
    else:
        test2()
            

if __name__ == '__main__':
    import random
    try:
        main()
    except KeyboardInterrupt:
        print('interrupted')



import re
from os.path import join, dirname, abspath

from firewall import util

public = util.public(globals())

_uescre = re.compile(r'\\(.)', re.S)

_uesctbl = dict(n = '\n', t = '\t')
_uesctbl['\n'] = ''
@public
def uescape(c):
    chr = c.group(1)
    if chr in _uesctbl:
        return _uesctbl[chr]
    return chr

_escre=re.compile(r'["\n\t\\]')
_esctbl = {'\n': 'n', '\t': 't'}
@public
def escape(c):
    chr = c.group(0)
    if chr in _esctbl:
        return "\\" + _esctbl[chr]
    return "\\" + chr


(WHITESPACE, CHAR, STRING, WORD) = range(4)

class Tokenizer(object):
    def __init__(self, strm):
        self.strm = strm
        
        self.buf = ''
        self.buflen = 0
        self.bufidx = 0
        self.lineno = 1
        
    wordchars = set('abcdefghijklmnopqrstuvwxyz'
                    'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                    '0123456789~@%^_+-=\\/:?*.,')
    
    wschars = set(" \t\r")

    def _getmore(self):
        ndat = self.strm.read(2048)
        if not ndat:
            return False
        
        self.buf = self.buf[self.bufidx:] + ndat
        self.bufidx = 0
        self.buflen = len(self.buf)
        return True

    def _read_while(self, func, sofs = 0, eofs = 0):
        src = self.buf
        len = self.buflen
        end = start = self.bufidx
        line = self.lineno

        lc = None
        while True:
            end += 1
            if end >= len:
                if not self._getmore(): break
                end = len - start; start = 0
                src = self.buf; len = self.buflen
                
            c = src[end]
                
                
            if lc == '\\' or func(c):
                lc = c
            else:
                break

            if c == '\n':
                line += 1

        ret = self.buf[start + sofs: end]
        self.bufidx = end + eofs
        self.lineno = line
        return ret
    
    def _handle_word(self):
        wordchars = self.wordchars
        ret = self._read_while(lambda c: c in wordchars)
        return WORD, _uescre.sub(uescape, ret)

    def _handle_dquot(self):
        ret = self._read_while(lambda c: c != '"', 1, 1)
        return STRING, _uescre.sub(uescape, ret)
    
    def _handle_squot(self):
        ret = self._read_while(lambda c: c != "'", 1, 1)
        return STRING, ret
    
    def _handle_comment(self):
        self._read_while(lambda c: c != '\n')
        return CHAR, '\n'

    def _handle_newline(self):
        self.bufidx += 1
        self.lineno += 1
        return CHAR, '\n'

    def _handle_whitespace(self):
        wschars = self.wschars
        self._read_while(lambda c: c in wschars)
        return WHITESPACE, ''

    def _handle_default(self):
        idx = self.bufidx
        self.bufidx = idx + 1
        return CHAR, self.buf[idx]

    def __iter__(self):
        handlers = self.handlers
        while True:
            if self.bufidx >= self.buflen:
                if not self._getmore():
                    break
            c = self.buf[self.bufidx]
            typ, val = handlers[ord(c)](self)
            if typ != WHITESPACE:
                yield typ, val

        
    handlers = [_handle_default] * 256
    for z in wordchars:
        handlers[ord(z)] = _handle_word
        
    for z in wschars:
        handlers[ord(z)] = _handle_whitespace
        
    del z
    
    handlers[ord('\n')] = _handle_newline
    handlers[ord('"')] = _handle_dquot
    handlers[ord("'")] = _handle_squot
    handlers[ord('#')] = _handle_comment

class Config(object):
    public = True
    
    def __init__(self):
        self.seen = set()

    def parse_file(self, path):
        path = path.resolve()
        if path in self.seen:
            return
        
        self.seen.add(path)
        
        with path.open('r', encoding='utf8') as fp:
            yield from self._parse_file(path, fp)

    def _parse_file(self, path, fp):
        brace_lvl = 0
        ccmd = []
        cmd_lineno = None
        tk = Tokenizer(fp)
        for typ, val in tk:
            #print typ, repr(val)
            if typ == CHAR:
                if brace_lvl == 0 and (val == '\n' or val == ';'):
                    if ccmd:
                        yield path, cmd_lineno, ccmd
                        ccmd = []
                    cmd_lineno = None
                    continue
                
                if val == '{':
                    brace_lvl += 1
                elif val == '}':
                    if brace_lvl > 0:
                        brace_lvl -= 1
            elif typ == WORD:
                if val == '\n':
                    continue
                
            if cmd_lineno is None:
                cmd_lineno = tk.lineno
            ccmd.append(val)

        if ccmd:
            yield filname, cmd_lineno, ccmd
                
    def parse_with_include(self, path):
        for cmd in self.parse_file(path):
            args = cmd[2]
            if len(args) >= 2 and args[0] == 'require':
                nfil = path.parent / args[1]
                yield from self.parse_with_include(nfil)
            else:
                yield cmd
                
    def reset(self):
        self.seen.clear()

util.update(globals())

if __name__ == '__main__':
    import sys
    cfg = Config()
    for fn, line, cmd in cfg.parse_with_include(sys.argv[1]):
        print(fn, line, cmd)






import traceback
import struct
import os
import socket
import asyncio

from .util import Result

RESP_OK = 0
ERR_UNKNOWN_EXCEPTION = 257
ERR_MALFORMED_COMMAND = 258
ERR_INVALID_MODULE = 259
ERR_INVALID_FUNCTION = 260
ERR_COMMAND_LOST = 261
ERR_ACCESS_DENIED = 262

def split_mod_func(args):
    modfunc = args[0].split('.', 1)
    module = modfunc[0]
    if len(modfunc) == 1:
        func = args[1]
        args = args[2:]
    else:
        func = modfunc[1]
        args = args[1:]
    return module, func, args

async def command_handler(writer, tag, args, handler):
    try:
        code, text = (await handler(args))
    except Exception as exc:
        code = ERR_UNKNOWN_EXCEPTION
        text = traceback.format_exc()
        
    data = text.encode('utf8')
    writer.write(struct.pack(">IIH", tag, code, len(data)) + data)
    asyncio.create_task(writer.drain())

async def command_conn(reader, writer, handler):
    try:
        while True:
            dat = await reader.readexactly(4)
            if len(dat) < 4:
                break
        
            tag, = struct.unpack(">I", dat)
            args = []
            while True:
                dat = await reader.readexactly(2)
                ln, = struct.unpack(">H", dat)
                if ln == 65535:
                    break
            
                dat = await reader.readexactly(ln)
                args.append(dat.decode('utf8', 'ignore'))
            
            asyncio.create_task(command_handler(writer, tag, args, handler))
    except asyncio.IncompleteReadError:
        pass
    finally:
        writer.close()

async def command_server(path, handler, mode):
    try:
        os.unlink(path)
    except:
        pass

    server = await asyncio.start_unix_server(lambda reader, writer: command_conn(reader, writer, handler), path)
    os.chmod(path, mode)
    return server

class Client:
    def __init__(self, path):
        self.path = path
        self.reader = None
        self.writer = None
        self.readtask = None
        self.next_tag = 42
        self.pending_commands = {}

    async def connect(self):
        self.reader, self.writer = await asyncio.open_unix_connection(self.path)

        self._resp_header = None

        self.readtask = asyncio.create_task(self.read_responses())

    async def read_responses(self):
        while True:
            try:
                dat = await self.reader.readexactly(10)
            except (asyncio.IncompleteReadError, asyncio.CancelledError):
                break

            tag, status, leng = struct.unpack(">IIH", dat)
            if leng:
                try:
                    stxt = await self.reader.readexactly(leng)
                except (asyncio.IncompleteReadError, asyncio.CancelledError):
                    break
            else:
                stxt = ""

            try:
                event = self.pending_commands[tag]
                del self.pending_commands[tag]
            except KeyError:
                continue

            event.set((status, stxt.decode('utf8', 'ignore')))

    async def send_command(self, *args):
        tag = self.next_tag
        self.next_tag += 1

        cmdcond = Result()
        self.pending_commands[tag] = cmdcond

        args = [arg.encode('utf8') for arg in args]
        
        data = (struct.pack(">I", tag) + b''.join(struct.pack(">H", len(s)) + s for s in args) +
               struct.pack(">H", 65535))

        self.writer.write(data)

        return await cmdcond.wait()
        
    def close(self):
        if self.readtask:
            self.readtask.cancel()

        self.readtask = None
        self.writer.close()
        self.reader = None
        self.writer = None
        self.socket = None
        
    def has_pending(self):
        return bool(self.pending_commands)
    



import sys
import re
import traceback
import asyncio
import inspect

import types

import os

from collections import deque
from .config import Config

from pathlib import Path

from firewall import util, command

public = util.public(globals())

modules = {}

SOURCE_USER, SOURCE_ADMIN, SOURCE_CONFIG = range(3)
MODULE_PREFIX = 'firewall.modules.'

__commands__ = {}

class Command(object):
    public = True
    def __init__(self, func, admin):
        self.func = func
        self.admin = admin

    async def call(self, source, *args, **kwargs):
        if self.admin and source == SOURCE_USER:
            return (command.ERR_ACCESS_DENIED, 'access denied')
        return await self.func(source, *args, **kwargs)
    
def _command(func, admin):
    mod = sys.modules[func.__module__]
    try:
        cmd_dict = mod.__commands__
    except:
        cmd_dict = mod.__commands__ = {}
        
    name = func.__name__
    if name.startswith('command_'):
        name = name[8:]

    cmd_dict[name] = Command(func, admin)
    return func

@public
def user_command(func):
    return _command(func, False)

@public
def admin_command(func):
    return _command(func, True)

@public
def hook(f):
    def hookfunc(mod, called, q, args, kwargs):
        return f(*args, **kwargs)
    hookfunc.__name__ = f.__name__
    return hookfunc
    
@public
def after(*deps):
    def ret(f):
        f.deps = deps
        return f
    return ret

async def _callhook(mods, hookname, args, kwargs):
    called = set()
    q = deque(iter(mods.items()))
    while q:
        name, mod = q.popleft()
        called.add(name)
        
        try:
            func = getattr(mod, hookname)
        except AttributeError:
            continue
        
        task = None
        deps = getattr(func, 'deps', None)
        if deps:
            for dep in deps:
                if dep in mods and not dep in called:
                    q.append((name, mod))
                    break
            else:
                task = func(*args, **kwargs)

        task = func(*args, **kwargs)

        if inspect.isawaitable(task):
            await task

@public
def callhook(name, *args, **kwargs):
    return _callhook(modules, name, args, kwargs)


def modmtime(module):
    mfil = module.__file__
    return Path(mfil).stat().st_mtime

def do_import(name):
    mod = __import__(name)
    for c in name.split('.')[1:]:
        mod = getattr(mod, c)
    return mod

@public
def insert_module(name, log=True):
    modname = MODULE_PREFIX + name
    if name in modules:
        return modules[name]
    
    loaded_already = (modname in sys.modules)
        
    try:
        module = do_import(modname)
        module.__modtime__ = modmtime(module)

        module.init()
        modules[name] = module
        if log:
            print('module %s initialized' % name, file=sys.stderr)
    except:
        if not loaded_already and modname in sys.modules:
            del sys.modules[modname]
        raise
    return module
    
@public
def remove_module(mod):
    modname = mod.__name__
    name = modname.split('.')[-1]
    

    if name in modules or modules[name] != mod:
        raise ValueError("unknown module")

    mod.cleanup()
    if name in modules:
        del modules[name]

    if modname in sys.modules:
        del sys.modules[modname]

@public
def reload_module(mod):
    name = mod.split('.')[-1]
    remove_module(mod)
    insert_module(name, log = False)
    print('module %s reloaded' % name, file=sys.stderr)
    
def check_reload(module, force = False):
    modtime = modmtime(module)
    omodtime = module.__modtime__
    if force or modtime > omodtime:
        return reload_module(module)
    
    return module


@user_command
def lsmod(source):
    return (command.RESP_OK, '\n'.join(iter(modules.keys())))

def command_handler(source):
    async def handle_command(args):
        modname, funcnm, args = command.split_mod_func(args)

        if modname == 'core':
            dct = __commands__
        else:
            try:
                dct = modules[modname].__commands__
            except (KeyError, AttributeError):
                return (command.ERR_INVALID_MODULE, 'invalid module: %s' % modname)

        try:
            func = dct[funcnm]
        except KeyError:
            return (command.ERR_INVALID_MODULE, 'invalid function: %s.%s' % (modname, funcnm))
            
                
        return await func.call(source, *args)
    return handle_command

user_command_handler = command_handler(SOURCE_USER)
admin_command_handler = command_handler(SOURCE_ADMIN)
config_command_handler = command_handler(SOURCE_CONFIG)


def gen_config_files(path):
    if path.is_dir():
        for f in path.iterdir():
            if f.name.startswith('.') or f.suffix != '.conf':
                continue
            
            yield f
    else:
        yield path

@public
async def reload_config(forcemods = []):
    if forcemods:
        if forcemods[0] == 'all':
            forcemods = list(modules.keys())
            
        for m in forcemods:
            remove_module(modules[m])

    await callhook('clear_config')
    seen_mods = set()

    cfg = Config()

    curmod = 'core'
    
    current_mods = set(modules.keys())
    for fn in gen_config_files(Path(_config_file_name)):
        for fname, line, cmd in cfg.parse_with_include(fn):
            try:
                nm = cmd[0]
                if nm == '[':
                    if len(cmd) != 3 or cmd[2] != ']':
                        raise ValueError('invalid section header')
                    curmod = cmd[1]
                    if curmod != 'core':
                        seen_mods.add(curmod)
                        if not curmod in modules:
                            insert_module(curmod)
                else:
                    modlst = nm.split('.', 1)
                    if len(modlst) == 1:
                        module = curmod
                        func = modlst[0]
                    else:
                        module, func = modlst
                    await config_command_handler([module, func] + cmd[1:])
            except Exception as exc:
                traceback.print_exc()
                
    dead_mods = current_mods - seen_mods
    for m in dead_mods:
        try:
            remove_module(modules[m])
        except KeyError:
            pass

    for m in modules.values():
        m.__dict__.update(modules)
    
    return await callhook('config_loaded')
    
@admin_command
async def command_reload(source, *forcemods):
    await reload_config(forcemods)
    return (command.RESP_OK, 'config reloaded')


async def run_server(server):
    async with server:
        await server.serve_forever()

async def run(config, asockpath, usockpath, notify=None):
    global _config_file_name, _config
    _config_file_name = config

    servers = []
    
    servers.append(await command.command_server(asockpath, admin_command_handler, 0o600))
    if usockpath:
        servers.append(await command.command_server(usockpath, user_command_handler, 0o666))

    if notify:
        notify()

    await reload_config()


    
    await asyncio.gather(*[run_server(server) for server in servers])

util.update(globals())
if __name__ == '__main__':
    sys.modules['firewall.core'] = sys.modules[__name__]
    mydir = Path(__file__).parent

    init(mydir / 'config', mydir / 'control', mydir / 'ucontrol')
    



import asyncio

__all__ = ['public', 'update']

class Result(asyncio.Event):
    def __init__(self):
        super().__init__()
        self._result = None

    def set(self, val=None):
        self._result = val
        super().set()

    async def wait(self):
        await super().wait()
        return self._result

def public(globs):
    all = globs['__all__'] = []
    def ret(f):
        all.append(f.__name__ if hasattr(f, '__name__') else f)
        return f
    return ret

def update(globs):
    all = globs['__all__']
    name = globs['__name__']
    for k, v in globs.items():
        if isinstance(v, type):
            if v.__module__ == name and hasattr(v, 'public') and v.public is True:
                all.append(k)
        if k.upper() == k and not k.startswith('_'):
            all.append(k)
            



import os
import time

from os.path import *

class Cache(object):
    def __init__(self, load=None):
        self.cache = {}
        if load is not None:
            self.load = load

    def get(self, key, *loadargs):
        obj = self.cache.get(key)
        nobj = self.load(key, obj, *loadargs)
        if nobj is not obj:
            if hasattr(obj, 'close'):
                obj.close()
            if nobj is None:
                try:
                    del self.cache[key]
                except KeyError:
                    pass
            else:
                self.cache[key] = nobj

        return nobj

    def load(self, key, obj):
        pass

class FileCache(Cache):
    def __init__(self, loadfile=None):
        Cache.__init__(self)
        if loadfile is not None:
            self.loadfile = loadfile

    def load(self, key, obj, stat=None):
        cachemtime = None
        if obj is not None:
            cachemtime = obj.stat.st_mtime

        if stat is None:
            try:
                stat = os.stat(key)
            except Exception:
                return

        if cachemtime is None or stat.st_mtime > cachemtime:
            return self.loadfile(key, stat)
        return obj

    def loadfile(self, path, stat):
        pass



import sys
import re
import asyncio

from io import StringIO

from firewall.core import *
from firewall.modules.iptables import *

nat_nets = []
forward = []
public_iface = None
enable_dnat_loopback = False

class NatForward(object):
    def __init__(self, toaddr, sport, dport = None, nports = 1, proto='tcp'):
        if dport is None:
            dport = sport

        self.toaddr = toaddr
        self.proto = proto
        self.sport = sport
        self.dport = dport
        self.nports = nports

    def gen_rule(self, snataddr=None):
        if self.nports == 1:
            oport = str(self.sport)
            toport = str(self.dport)
        else:
            oport = '%d:%d' % (self.sport, self.sport + self.nports - 1)
            toport = '%d-%d' % (self.dport, self.dport + self.nports - 1)
        if snataddr:
            ruletxt = '-p %s -d %s --dport %s --to-source %s' % (self.proto, self.toaddr, toport.replace('-', ':'), snataddr)
            return Rule(ruletxt, 'SNAT')
        else:
            ruletxt = '-p %s --dport %s --to-destination %s:%s' % (self.proto, oport, self.toaddr, toport)
            return Rule(ruletxt, 'DNAT')

def fill_tables(filter, nat, mangle, raw):
    if not public_iface:
        return
    try:
        pubnet = get_network(public_iface)
        pubaddr = pubnet.getaddr()
    except NetworkUnavail as e:
        print("public interface not found: %s" % e, file=sys.stderr)
        return

    natout = nat['OUTPUT']
    natpost = nat['POSTROUTING']
    natpre = nat['PREROUTING']

    natmasq = nat['MASQ']

    portforward = nat['PORTFORWARD']

    natpre.add(Rule('--destination %s' % pubaddr, 'PORTFORWARD'))
    natout.add(Rule('--destination %s' % pubaddr, 'PORTFORWARD'))

    for f in forward:
        natmasq.add(f.gen_rule(pubaddr))

    for n in nat_nets:
        try:
            net = get_network(n).getnet()
            natmasq.add(Rule('--destination %s' % net, 'RETURN'))
            if enable_dnat_loopback:
                natpost.add(Rule('--source %s' % (net), 'MASQ'))
            else:
                natpost.add(Rule('--source %s --out-interface %s' % (net, pubnet.iface), 'MASQ'))
        except NetworkUnavail as e:
            pass
            #print >> sys.stderr, "interface not found: %s" % e

    natmasq.add(Rule('--to-source %s' % pubaddr, 'SNAT'))

    for f in forward:
        portforward.add(f.gen_rule())

def clear_config():
    global public_iface, enable_dnat_loopback
    public_iface = None
    enable_dnat_loopback = False
    del nat_nets[:]
    del forward[:]

@admin_command
async def masqnet(source, *args):
    for n in args:
        nat_nets.append(n)
    return await reload_tables_for_cmd(source, 'ok')

@admin_command
async def masqto(source, net):
    global public_iface
    public_iface = net
    return await reload_tables_for_cmd(source, 'ok')

@admin_command
async def fwd(source, addr, port, toport, nports, proto='tcp'):
    port = int(port)
    toport = int(toport)
    nports = int(nports)
    forward.append(NatForward(addr, port, toport, nports, proto))
    return await reload_tables_for_cmd(source, 'ok')

@admin_command
async def dnat_loopback(source, enable):
    global enable_dnat_loopback
    enable_dnat_loopback = enable.lower() in ('yes', 'true', '1')
    return await reload_tables_for_cmd(source, 'ok')

def init():
    pass

def cleanup():
    pass



import sys
import re
import socket
from firewall import util
from firewall.getifaddrs import *

public = util.public(globals())

class Token(object):
    def __init__(self, pat):
        self.rx = re.compile(pat)
        
    def match(self, txt, pos):
        m = self.rx.match(txt, pos)
        if m:
            return self.getval(m), m.end()
        return None, None

    def getval(self, m):
        grp = m.groups()
        if grp:
            return grp[0]
    def __repr__(self):
        return '<tk /%s/>' % self.rx.pattern
public(Token)


class AddrExpr(object):
    def eval(self, ctx):
        pass
    
    def filter(self, ctx, addr):
        for myaddr in self.eval(ctx):
            if myaddr.family == addr.family:
                mask = myaddr.mask
                if myaddr.addr & mask == addr.addr & mask:
                    return True
        return False
    
    def get_iface(self, ctx):
        l = self.eval(ctx)
        if l:
            return l[0].iface or None
        else:
            return None

class AddrLiteral(AddrExpr):
    def __init__(self, v):
        self.v = v

    def eval(self, ctx):
        return [self.v]

    def __repr__(self):
        return '%s/%d' % (self.v.getaddr(), self.v.bits)

class AliasExpr(AddrExpr):
    def __init__(self, name):
        self.name = name

    def eval(self, ctx):
        try:
            a = ctx.getalias(self.name)
        except KeyError:
            return []
        return a.eval(ctx)

    def filter(self, ctx, addr):
        try:
            a = ctx.getalias(self.name)
        except KeyError:
            return False
        return a.filter(ctx, addr)
    
    def __repr__(self):
        return self.name
    
class IfaceExpr(AddrExpr):
    def __init__(self, name, mask=None):
        self.name = name
        self.mask = None

    def getaddrs(self):
        if self.mask is None:
            return ctx.getifaceaddrs(self.name)
        addrs = [a.copy() for a in ctx.getifaceaddrs(self.name)]
        for a in addrs:
            a.mask = self.mask

    def eval(self, ctx):
        return ctx.getifaceaddrs(self.name)

    def __repr__(self):
        return '@' + self.name
    
class BinOp(AddrExpr):
    rtassoc = False
    postop = False
    prec = 3
    op = 'null'
    def __init__(self, l, r=None):
        self.l = l
        self.r = r

    def __repr__(self):
        if self.postop:
            return '(%r %s)' % (self.l, self.op)
        else:
            return '(%r %s %r)' % (self.l, self.op, self.r)

nullop = BinOp(None)
nullop.prec = -1

class SetInterface(AddrExpr):
    postop = True
    prec = 0
    def __init__(self, iface, l=None):
        self.l = l
        self.iface = iface
    def __repr__(self):
        return '(%r @%s)' % (self.l, self.iface)

    def eval(self, ctx):
        return [a.copy(iface=self.iface) for a in self.l.eval(ctx)]

    def __call__(self, l, r):
        return SetInterface(self.iface, l)
    
    def get_iface(self, ctx):
        return self.iface
class LogOr(BinOp):
    prec = 1
    op = '|'

    def eval(self, ctx):
        return self.l.eval(ctx) + self.r.eval(ctx)

    def filter(self, ctx, addr):
        return self.l.filter(ctx, addr) or self.r.filter(ctx, addr)

    
class Merge(LogOr):
    prec = 3
    op = '+'

    def eval(self, ctx):
        aa = self.l.eval(ctx)
        ba = self.r.eval(ctx)
        ret = []
        for a in aa:
            for b in ba:
                if a.family != b.family:
                    continue
                na = (a.addr & a.mask) | (b.addr & ~a.mask)
                ret.append(NetAddr(a.iface, a.family, na, a.mask | b.mask, a.scope))
        return ret

class LogAnd(LogOr):
    rtassoc = True
    prec = 2
    op = '&'
    def eval(self, ctx):
        return [a for a in self.l.eval(ctx) if self.r.filter(ctx, a)]

    def filter(self, ctx, addr):
        return self.l.filter(ctx, addr) and self.r.filter(ctx, addr)

class Bracket(LogAnd):
    rtassoc = False
    postop = True
    def __init__(self, r):
        self.l = nullop
        self.r = r
    
    def __call__(self, l, r):
        return LogAnd(l, self.r)

class LogNot(LogOr):
    op = '!'
    def eval(self, ctx):
        return self.l.eval(ctx)

    def filter(self, ctx, addr):
        return not self.l.filter(ctx, addr)

    def __repr__(self):
        return '(%s%r)' % (self.op, self.l)


tkiface = Token(r'\s*@\s*([_a-zA-Z][-a-zA-Z0-9]*\+?)')
tkalias = Token(r'\s*([_a-zA-Z][-a-zA-Z0-9]*)')

# _not_ meant to check for valid address!
tkipv6 = Token(r'\s*([0-9a-fA-F]*:[0-9a-fA-F:.]+)')
tkipv4 = Token(r'\s*([0-9.]+)')
tkchar = Token(r'\s*([,\|&\[\]\(\)\!+])')
tknetmask = Token(r'\s*/\s*(\d+)')

class AddrParser(object):
    maxlvl = 3
    
    def __init__(self, txt, pos=0):
        self.txt = txt
        self.pos = pos
        self.lpos = None

    def read(self, *types):
        self.lpos = self.pos
        for tk in types:
            v, e = tk.match(self.txt, self.pos)
            if e is not None:
                self.pos = e
                return tk, v
        return None, None

    def back(self):
        self.pos = self.lpos

    def isend(self):
        return self.pos >= len(self.txt)

    def error(self):
        #print 'error at %d %r' % (self.pos, self.txt[self.pos:])
        raise ValueError('invalid expression')

    def getalias(self, alias):
        try:
            return self.aliases[alias]
        except KeyError:
            raise NetworkUnavail

    def make_addr(self, family, val, size):
        mask = (1 << size) - 1
        tk, v = self.read(tknetmask)
        if v:
            mask &= ~((1 << (size - int(v))) - 1)
        else:
            self.back()
        
        return AddrLiteral(NetAddr('', family, val, mask, 0))

    def rest(self):
        return self.txt[self.pos:]
    
    def read_op(self):
        tk, v = self.read(tkchar, tkiface)
        if tk is None:
            return nullop
        
        if tk == tkchar:
            if v == '|' or v == ',':
                return LogOr
            if v == '&':
                return LogAnd
            if v == '+':
                return Merge
            if v == '[':
                self.back()
                l = self.parse_atom()
                return Bracket(l)
        elif tk == tkiface:
            return SetInterface(v)
        return nullop
    
    def parse_atom(self):
        tk, v = self.read(tkipv6, tkalias, tkiface, tkipv4, tkchar)
        if tk == tkalias:
            return AliasExpr(v)
        if tk == tkiface:
            return IfaceExpr(v)
        if tk == tkipv4:
            dots = v.count('.')
            if dots > 3:
                self.error()
            elif dots < 3:
                v += '.0' * (3 - dots)
            return self.make_addr(socket.AF_INET, v, 32)
        if tk == tkipv6:
            return self.make_addr(socket.AF_INET6, v, 128)
        if tk == tkchar:
            if v == '!':
                return LogNot(self.parse_atom())
            if v == '(' or v == '[':
                e = self.parse_expr()
                tk, nv = self.read(tkchar)
                if (v == '(' and nv != ')') or (v == '[' and nv != ']'):
                    self.error()
                return e
        self.error()
        
    def parse_expr(self, minlvl=0, maxlvl=None):
        if maxlvl is None:
            maxlvl = self.maxlvl

        clr = None, None, None
        maxlvl -= minlvl
        stk = [clr] * (maxlvl)
        clvl = maxlvl
        lf = self.parse_atom()

        ret = prev = op = lastop = None
        while True:
            if op:
                if prev and op.rtassoc and prev.r:
                    prev.r = op(prev.r, ret)
                    prev = prev.r
                else:
                    prev = lf = op(lf, ret)

            op = lastop or self.read_op()
            rt_prec = op.prec - minlvl
            if rt_prec != clvl:
                if clvl:
                    lastop = op
                    ret = lf
                    clvl -= 1
                    prev, op, lf = stk[clvl]
                    if not lf:
                        lf = ret
                    stk[clvl] = clr
                    continue
                else:
                    self.back()
                    #self.pos = self.op_start
                    return lf
            lastop = None
            
            if op.postop:
                ret = None
                clvl = maxlvl
                continue
            if clvl == maxlvl:
                ret = self.parse_atom()
            else:
                stk[clvl] = prev, op, lf
                prev = op = None
                lf = self.parse_atom()
                clvl = maxlvl
                continue
public(AddrParser)

class EvalContext(object):
    def __init__(self):
        self.aliases = {}
        self.ifaces = {}

    def getalias(self, name):
        return self.aliases[name]

    def getifaceaddrs(self, name):
        return self.ifaces[name]

    def updateifaces(self):
        self.ifaces = getifaddrs()
public(EvalContext)



import sys
import time
import os
import re
import asyncio

import random
import socket

from os.path import getmtime, dirname, join, abspath

from io import StringIO

from firewall import traflog
from firewall.core import *
from firewall.modules.iptables import *
from firewall.modules.manual import filter_args

class IptCounter(object):
    def __init__(self, ctrs, ruleargs):
        self.lastv = 0
        self.ctrs = ctrs
        self.ruleargs = ruleargs

    def update(self, nv):
        delta = nv - self.lastv
        if delta < 0:
            delta = nv
        #if delta > 3000000 * 5:
        #    print 'big delta! %d %d %d' % (delta, nv, self.lastv)
        self.lastv = nv
        for ctr in self.ctrs:
            ctr.delta += delta

CHAINBASE = "LOG_"
class TrafMon(object):
    def __init__(self):
        self.logdir = '/var/log/traflog'
        self.iptables_bin = '/sbin/iptables'
        self.interval = 5
        self.counters = {}
        self.iptctrs = []
        self.chainname = None
        self._update_task = None
        self._run_task = None
        
    def addcounter(self, name, ctrnames):
        path = join(self.logdir, name)
        ctr = traflog.TrafLog(path, len(ctrnames))
        ctr.seek_end()
        ctr.close()
        
        self.counters[name] = ctr

    async def _update_counters(self):
        if not self.chainname:
            return

        #print 'updating from %s' % (self.chainname)
        proc = await asyncio.create_subprocess_exec(self.iptables_bin, '-L', self.chainname, '-nxv', stdout=subprocess.PIPE)
        stdout, stderr = await proc.communicate()
        ret = await proc.wait()

        lines = stdout.decode('utf8', 'ignore').splitlines()

        iptitr = iter(self.iptctrs)
        for ln in lines:
            m = rxbcount.match(ln)
            if m:
                try:
                    nxt = next(iptitr)
                    nxt.update(int(m.group(1)))
                except StopIteration:
                    pass
                
        ctime = traflog.gettime()
        for ctr in self.counters.values():
            ctr.write_counters(ctime)
        #print 'done updating'
        
    async def update_counters(self):
        #print 'starting update'
        if not self._update_task or self._update_task.done():
            self._update_task = asyncio.create_task(self._update_counters())
            
        #print 'waiting for update'
        await self._update_task
        #print 'done waiting'

    
    def new_chain(self):
        for i in self.iptctrs:
            i.lastv = 0
        ccn = self.chainname
        ncn = CHAINBASE + '%08X' % random.randrange(0, 1<<32)
        self.chainname = ncn
        #print 'new chain: %s -> %s' % (ccn, ncn)
        return ncn
        

    async def _run_update(self):
        while True:
            await self.update_counters()
            await asyncio.sleep(self.interval)

    def start(self):
        if not self._run_task:
            self._run_task = asyncio.create_task(self._run_update())
        
    def stop(self):
        if self._run_task:
            self._run_task.cancel()
            self._run_task = None
        
        for ctr in self.counters.values():
            ctr.close()
            
rxbcount = re.compile(r'\s*\d+\s+(\d+)')

monitor = TrafMon()
old_monitor = None

@admin_command
async def command_interval(source, tstr):
    monitor.interval = int(tstr)
    return (0, 'ok')

@admin_command
async def command_iptables_bin(source, tstr):
    monitor.iptables_bin = tstr
    return (0, 'ok')

@admin_command
async def command_logdir(source, tstr):
    monitor.logdir = tstr
    return (0, 'ok')

@admin_command
async def command_counter(source, name, *ctrnames):
    monitor.addcounter(name, ctrnames)
    return (0, 'ok')

@admin_command
async def command_rule(source, *args):
    itr = iter(args)
    ctrs = []
    for r in itr:
        if r == '--':
            break
        ctr, s, rn = r.partition('.')
        if s:
            rn = int(rn)
        else:
            ctr = r
            rn = 0
        ctrs.append(monitor.counters[ctr].ctrs[rn])
    
    monitor.iptctrs.append(IptCounter(ctrs, list(itr)))
    return (0, 'ok')

@asyncfunc
async def fill_tables(filter, nat, mangle, raw):
    await monitor.update_counters()
    chain = monitor.new_chain()
    
    filter_log = filter[chain]
    
    for iptc in monitor.iptctrs:
        ruletxt = ' '.join(filter_args(iptc.ruleargs))
        filter_log.add(Rule(ruletxt, 'NOOP'))
        
    noop = filter['NOOP']
    filter['INPUT'].add(Rule('', chain))
    filter['OUTPUT'].add(Rule('', chain))
    filter['FORWARD'].add(Rule('', chain))
    #raw['PREROUTING'].add(Rule('-p tcp -m tcp --dport 22 --tcp-flags SYN,RST,ACK SYN', CHAINNAME))

def init():
    pass

def cleanup():
    monitor.stop()
    
async def clear_config():
    global monitor, old_monitor
    await monitor.update_counters()
    
    monitor.stop()

    old_monitor = monitor
    monitor = TrafMon()
    
def config_loaded():
    global old_monitor
    monitor.start()

    old_monitor = None




import sys
import time
import os
import re
import asyncio

import socket

from os.path import getmtime, dirname, join, abspath

from io import StringIO

from firewall.core import *
from firewall.modules.iptables import *

class Options(object):
    def __init__(self):
        self.failcnt = 4
        self.bantime = 30 * 60
        self.exempt_addrs = []

CHAINNAME = "SSHkill"

options = Options()
old_options = None
failcnt = {}
bantimers = {}

class BanTimer(object):
    def __init__(self, btime, addr):
        self.addr = addr
        self.endtime = time.time() + btime
        self.time = btime
        self.task = asyncio.create_task(self.unban_timer())
        
        if addr in failcnt:
            del failcnt[addr]
        
    async def unban_timer(self):
        await asyncio.sleep(self.time)
        print("SSH ban expired for address %s" % self.addr, file=sys.stderr)
        try:
            del bantimers[self.addr]
        except KeyError:
            pass
        return await reload_tables()
        
    def cancel(self):
        self.task.cancel()

def ban(addr):
    #reset the counter
    bantimers[addr] = BanTimer(options.bantime, addr)

    asyncio.create_task(reload_tables())

def badlogin(addr):
    iaddr = iptables.pack_addr(addr)
    for ex in options.exempt_addrs:
        try:
            net = get_network(ex)
            a = net._getaddr()
            m = net._getmask()
            if iaddr & m == a & m:
                return
        except NetworkUnavail:
            pass

    if addr in bantimers:
        # Address already banned -- this can happen since sshd gives
        # the user three tries before disconnecting, and the filter
        # only stops new connections.
        return
    
    cnt = failcnt.get(addr, 0)
    cnt += 1
    failcnt[addr] = cnt
    
    if cnt >= options.failcnt:
        print("%d failed logins from %s, banning for %d seconds" % (cnt, addr, options.bantime), file=sys.stderr)
        ban(addr)
    else:
        print('%d out of %d failed logins from %s' % (cnt, options.failcnt, addr), file=sys.stderr)

def goodlogin(addr):
    reload = False
    try:
        del bantimers[addr]
        reload = True
    except KeyError:
        pass
    
    try:
        del failcnt[addr]
    except KeyError:
        pass

    if reload:
        reload_tables().start()

def fill_tables(filter, nat, mangle, raw):
    filter_sshkill = filter[CHAINNAME]
    #raw_sshkill = raw[CHAINNAME]
    
    for tmr in bantimers.keys():
        filter_sshkill.add(Rule('-p tcp -s %s' % tmr, 'DROP'))

    filter['INPUT'].add(Rule('-p tcp -m tcp --dport 22 --tcp-flags SYN,RST,ACK SYN', CHAINNAME))
    #raw['PREROUTING'].add(Rule('-p tcp -m tcp --dport 22 --tcp-flags SYN,RST,ACK SYN', CHAINNAME))

@admin_command
async def command_bantime(source, tstr):
    options.bantime = parsetime(tstr)
    return (0, 'ok')

@admin_command
async def command_exempt(source, *addrs):
    options.exempt_addrs.extend(addrs)
    return (0, 'ok')

@admin_command
async def command_failcnt(source, n):
    options.failcnt = int(n)
    return (0, 'ok')

@admin_command
async def command_badlogin(source, addr, user):
    badlogin(addr)
    return (0, 'ok')
    
@admin_command
async def command_goodlogin(source, addr, user):
    goodlogin(addr)
    return (0, 'ok')

@admin_command
async def command_force(source, addr, timestr):
    bantime = parsetime(timestr)
    print('SSH ban forced for %s, %d seconds' % (addr, bantime), file=sys.stderr)
    try:
        bantimers[addr].cancel()
    except KeyError:
        pass
    
    bantimers[addr] = BanTimer(bantime, addr)
    return await reload_tables_for_cmd(source, 'ban successfully added')

@admin_command
async def command_cancel(cmd, addr):
    try:
        bantimers[addr].cancel()
        del bantimers[addr]
        print("SSH ban canceled for address %s" % addr, file=sys.stderr)
        return await reload_tables_for_cmd(source, 'ban canceled')
    except KeyError:
        return (2, 'address was not banned')
    
@user_command
async def command_list(source):
    ctime = time.time()
    resp = StringIO()
    for v in list(bantimers.values()):
        resp.write('%s\t%d\n'%(v.addr, int(v.endtime - ctime)))

    return (0, resp.getvalue())
    
@admin_command    
async def command_clear(source):
    print('SSH bans cleared', file=sys.stderr)
    for v in list(bantimers.values()):
        v.cancel()
        
    bantimers.clear()
    failcnt.clear()
    return await reload_tables_for_cmd(source, 'bans cleared')


def init():
    clear_config()

def cleanup():
    for v in list(bantimers.values()):
        v.cancel()
        
    bantimers.clear()

def clear_config():
    global options, old_options
    old_options = options
    options = Options()

def config_loaded():
    global old_options
    old_options = None

